/*!
    \class MSpinner
    \brief MSpinner widget provides a spinning busy indicator.
    \inmodule MWidgets

    The MSpinner widget provides a spinning busy indicator.
*/

#include "mspinner.h"

#include <QTimer>
#include <QPainter>

struct MSpinnerPrivate {
    MSpinnerPrivate() :
        timer(new QTimer),
        current_frame(0),
        colorize(false)
    {}

    QTimer timer;
    int current_frame;
    bool colorize;

    int side;
    int dot_size;
    int dot_size_half;
    int dot_radius;
};

MSpinner::MSpinner(QWidget *parent) :
    QWidget(parent),
    d_ptr(new MSpinnerPrivate)
{
    connect(&d_ptr->timer, &QTimer::timeout, [=]() {
        if (d_ptr->current_frame >= 31) {
            d_ptr->current_frame = 0;
        } else {
            d_ptr->current_frame++;
        }
        update();
    });
    d_ptr->timer.setInterval(25);
}

MSpinner::~MSpinner()
{

}


void MSpinner::paintEvent(QPaintEvent *event)
{
    QWidget::paintEvent(event);

    QPainter painter(this);

    painter.setRenderHint(QPainter::Antialiasing);
    painter.translate(width() / 2, height() / 2);

    painter.setPen(Qt::NoPen);
    QBrush brush;
    if (d_ptr->colorize) {
        brush = palette().highlight();
    } else {
        brush = palette().text();
    }
    QColor color = brush.color();

    painter.rotate(-45.0);
    painter.rotate((d_ptr->current_frame/4) * 45.0);
    float alpha = (d_ptr->current_frame%4) * 0.025f;

    for (int i = 0; i < 8; ++i) {
        color.setAlphaF(qBound(0.0f, (i/8.0f)+0.2f-alpha, 1.0f));
        brush.setColor(color);
        painter.setBrush(brush);
        painter.drawEllipse(d_ptr->dot_radius, d_ptr->dot_size_half, d_ptr->dot_size, d_ptr->dot_size);
        painter.rotate(45.0);
    }
}

void MSpinner::setVisible(bool visible)
{
    if (visible) {
        d_ptr->current_frame = 0;
        d_ptr->timer.start();
    } else {
        d_ptr->timer.stop();
    }
    QWidget::setVisible(visible);
}

QSize MSpinner::sizeHint() const
{
    return minimumSizeHint();
}


QSize MSpinner::minimumSizeHint() const
{
    return QSize(16, 16);
}


void MSpinner::resizeEvent(QResizeEvent *event)
{

    d_ptr->side = qMin(width(), height());
    d_ptr->dot_size = d_ptr->side / 5;
    d_ptr->dot_size_half = -(d_ptr->dot_size / 2);
    d_ptr->dot_radius = (d_ptr->side/2) - d_ptr->dot_size - 2;

    QWidget::resizeEvent(event);
}

/*!
    \property MSpinner::colorize
    \brief whether the dots are colored

    This property describes whether the dots are painted with the highlight
    color. When disabled, the dots are painted in gray.
 */
bool MSpinner::colorize() const
{
    return d_ptr->colorize;
}

void MSpinner::setColorize(bool colorize)
{
    if (d_ptr->colorize == colorize)
        return;

    d_ptr->colorize = colorize;
    emit colorizeChanged(colorize);
}

