#include "settings.h"

// Qt
#include <QSettings>

#include <MUnixAutostart>
#include <MUnixApplicationIcon>

Settings *Settings::m_instance = nullptr;

Settings::Settings(QObject *parent) : QObject(parent)
{
    if (m_instance != nullptr) {
        qWarning("Only one instance of Settings allowed");
        return;
    }
    m_instance = this;
    load();
}

Settings::~Settings()
{
    save();
    m_instance = nullptr;
}

Settings &Settings::instance()
{
    return *m_instance;
}

void Settings::load()
{
    QSettings qsettings;
    qsettings.setIniCodec("UTF-8");
    qsettings.beginGroup(QLatin1String("Settings"));
    setTrayIconStashGroups(qsettings.value(QLatin1String("trayIconStashGroups"), true).toBool());
    setTrayIconEnabled(qsettings.value(QLatin1String("trayIconEnabled"), true).toBool());
    setTrayIconStartHidden(qsettings.value(QLatin1String("trayIconStartHidden"), true).toBool());
    setManagerShowStashGroups(qsettings.value(QLatin1String("managerShowStashGroups"), false).toBool());
    setUnmountAllOnExit(qsettings.value(QLatin1String("unmountAllOnExit"), true).toBool());
    setFileManagerCommand(qsettings.value(QLatin1String("fileManagerCommand"), QLatin1String("xdg-open \"$MOUNT_DIR\"")).toString());
    setMountAddedComand(qsettings.value(QLatin1String("mountAddedComand"), QLatin1String("")).toString());
    setMountRemovedCommand(qsettings.value(QLatin1String("mountRemovedCommand"), QLatin1String("")).toString());
    setBaseDir(qsettings.value(QLatin1String("baseDir"), QLatin1String("$HOME/Encfs")).toString());
    setStashGroups(qsettings.value(QLatin1String("stashGroups"), QStringList() << tr("Applications") << tr("Cloud") << tr("Removable") << tr("Others")).toStringList());
    setShowSidebar(qsettings.value(QLatin1String("showSidebar"), false).toBool());
    setIsFirstRun(qsettings.value(QLatin1String("isFirstRun"), true).toBool());
}
void Settings::save() const
{
    QSettings qsettings;
    qsettings.setIniCodec("UTF-8");
    qsettings.beginGroup(QLatin1String("Settings"));
    qsettings.setValue(QLatin1String("trayIconStashGroups"), trayIconStashGroups());
    qsettings.setValue(QLatin1String("trayIconEnabled"), trayIconEnabled());
    qsettings.setValue(QLatin1String("trayIconStartHidden"), trayIconStartHidden());
    qsettings.setValue(QLatin1String("managerShowStashGroups"), managerShowStashGroups());
    qsettings.setValue(QLatin1String("unmountAllOnExit"), unmountAllOnExit());
    qsettings.setValue(QLatin1String("fileManagerCommand"), fileManagerCommand());
    qsettings.setValue(QLatin1String("mountAddedComand"), mountAddedComand());
    qsettings.setValue(QLatin1String("mountRemovedCommand"), mountRemovedCommand());
    qsettings.setValue(QLatin1String("baseDir"), baseDir());
    qsettings.setValue(QLatin1String("stashGroups"), stashGroups());
    qsettings.setValue(QLatin1String("showSidebar"), showSidebar());
    qsettings.setValue(QLatin1String("isFirstRun"), isFirstRun());
}

bool Settings::trayIconStashGroups() const { return m_trayIconStashGroups; }
void Settings::setTrayIconStashGroups(bool trayIconStashGroups)
{
    if (m_trayIconStashGroups == trayIconStashGroups) {
        return;
    }

    m_trayIconStashGroups = trayIconStashGroups;
    emit trayIconStashGroupsChanged(trayIconStashGroups);
}

bool Settings::trayIconEnabled() const { return m_trayIconEnabled; }
void Settings::setTrayIconEnabled(bool trayIconEnabled)
{
    if (m_trayIconEnabled == trayIconEnabled) {
        return;
    }

    m_trayIconEnabled = trayIconEnabled;
    emit trayIconEnabledChanged(trayIconEnabled);
}

bool Settings::trayIconStartHidden() const { return m_trayIconStartHidden; }
void Settings::setTrayIconStartHidden(bool trayIconStartHidden)
{
    if (m_trayIconStartHidden == trayIconStartHidden)
        return;

    m_trayIconStartHidden = trayIconStartHidden;
    emit trayIconStartHiddenChanged(trayIconStartHidden);
}

bool Settings::managerShowStashGroups() const { return m_managerShowStashGroups; }
void Settings::setManagerShowStashGroups(bool managerShowStashGroups)
{
    if (m_managerShowStashGroups == managerShowStashGroups) {
        return;
    }

    m_managerShowStashGroups = managerShowStashGroups;
    emit managerShowStashGroupsChanged(managerShowStashGroups);
}

bool Settings::unmountAllOnExit() const { return m_unmountAllOnExit; }
void Settings::setUnmountAllOnExit(bool unmountAllOnExit)
{
    if (m_unmountAllOnExit == unmountAllOnExit)
        return;

    m_unmountAllOnExit = unmountAllOnExit;
    emit unmountAllOnExitChanged(unmountAllOnExit);
}

QString Settings::fileManagerCommand() const { return m_fileManagerCommand; }
void Settings::setFileManagerCommand(const QString &fileManagerCommand)
{
    if (m_fileManagerCommand == fileManagerCommand)
        return;

    m_fileManagerCommand = fileManagerCommand;
    emit fileManagerCommandChanged(fileManagerCommand);
}

QString Settings::mountAddedComand() const { return m_mountAddedComand; }
void Settings::setMountAddedComand(const QString &mountAddedComand)
{
    if (m_mountAddedComand == mountAddedComand)
        return;

    m_mountAddedComand = mountAddedComand;
    emit mountAddedComandChanged(mountAddedComand);
}

QString Settings::mountRemovedCommand() const { return m_mountRemovedCommand; }
void Settings::setMountRemovedCommand(const QString &mountRemovedCommand)
{
    if (m_mountRemovedCommand == mountRemovedCommand)
        return;

    m_mountRemovedCommand = mountRemovedCommand;
    emit mountRemovedCommandChanged(mountRemovedCommand);
}

QString Settings::baseDir() const { return m_baseDir; }
void Settings::setBaseDir(const QString &baseDir)
{
    if (m_baseDir == baseDir)
        return;

    m_baseDir = baseDir;
    emit baseDirChanged(baseDir);
}

QStringList Settings::stashGroups() const { return m_stashGroups; }
void Settings::setStashGroups(const QStringList &stashGroups)
{
    if (m_stashGroups == stashGroups) {
        return;
    }

    m_stashGroups = stashGroups;
    emit stashGroupsChanged(stashGroups);
}

bool Settings::showSidebar() const { return m_showSidebar; }
void Settings::setShowSidebar(bool showSidebar)
{
    if (m_showSidebar == showSidebar)
        return;

    m_showSidebar = showSidebar;
    emit showSidebarChanged(showSidebar);
}

bool Settings::isFirstRun() const { return m_isFirstRun; }
void Settings::setIsFirstRun(bool isFirstRun)
{
    if (m_isFirstRun == isFirstRun)
        return;

    m_isFirstRun = isFirstRun;
    emit isFirstRunChanged(isFirstRun);
}

bool Settings::isApplicationIconEnabled() const {
    return MUnixApplicationIcon().isEnabled();
}

void Settings::setApplicationIconEnabled(bool isApplicationIconEnabled) {
    MUnixApplicationIcon().setEnabled(isApplicationIconEnabled);

    emit isApplicationIconEnabledChanged(this->isApplicationIconEnabled());
}


bool Settings::isAutostartEnabled() const {
    return MUnixAutostart().isEnabled();
}

void Settings::setIsAutostartEnabled(bool isAutostartEnabled)
{
    MUnixAutostart().setEnabled(isAutostartEnabled);

    emit isAutostartEnabledChanged(this->isAutostartEnabled());
}
