#include "stashmanager.h"

// Qt
#include <QCoreApplication>
#include <QDebug>
#include <QSettings>
#include <QTimer>

#include "helpers.h"
#include <MUnixMountSurveiler>
#include "settings.h"
#include "stash.h"

StashManager *StashManager::m_instance = nullptr;

StashManager::StashManager(QObject *parent) :
    QObject(parent),
    m_mutex(),
    m_allStashes(),
    m_connections(),
    m_filteredStashes(),
    m_mountSurveiler(new MUnixMountSurveiler)
{
    if (m_instance != nullptr) {
        qDebug() << "Only one instance of StashManager is allowed.";
        return;
    }
    m_instance = this;

    connect(m_mountSurveiler.data(), &MUnixMountSurveiler::mountsChanged, [this]() {
        // Add a small delay for mtab to be updated by the kernel after the udev event
        QTimer::singleShot(200, [this]() {
            refresh();
        });
    });

    connect(m_mountSurveiler.data(), &MUnixMountSurveiler::mountsChanged, [this]() {
        // Add a another delay for removable devices because they might need some time to initialize
        QTimer::singleShot(5000, [this]() {
            refresh();
        });
    });

    connect(this, &StashManager::stashGroupChanged, this, &StashManager::rebuildFiltered);
    connect(&Settings::instance(), &Settings::stashGroupsChanged, this, &StashManager::rebuildFiltered);

    connect(this, &StashManager::deleteFinished, [this](const QSharedPointer<Stash> &stash, bool success, const QString &message) {
        Q_UNUSED(message)
        if (success) {
            removeStash(const_cast<QSharedPointer<Stash>&>(stash));
        }
    });

    connect(this, &StashManager::createFinished, [this](const QSharedPointer<Stash> &stash, bool success, const QString &message) {
        Q_UNUSED(message)
        if (success) {
            stash->setIsInitialized(true);
            save();
        }
    });

    load();
}

StashManager::~StashManager()
{
    save();

    foreach (auto stash, m_allStashes) {
        stash->disconnect();
    }

    m_instance = nullptr;
}

StashManager &StashManager::instance()
{
    return *m_instance;
}

QSharedPointer<Stash> StashManager::addStash(const QSharedPointer<Stash> &stash)
{
    m_mutex.lock();
    stash->refresh();
    stash->m_weakref = stash.toWeakRef();

    connect(stash.data(), &Stash::mountFinished, [this,stash](bool success, const QString &message) {
        emit mountFinished(stash, success, message);
    });
    connect(stash.data(), &Stash::unmountFinished, [this,stash](bool success, const QString &message) {
        emit unmountFinished(stash, success, message);
    });
    connect(stash.data(), &Stash::createFinished, [this,stash](bool success, const QString &message) {
        emit createFinished(stash, success, message);
    });
    connect(stash.data(), &Stash::changePasswordFinished, [this,stash](bool success, const QString &message) {
        emit changePasswordFinished(stash, success, message);
    });
    connect(stash.data(), &Stash::deleteFinished, [this,stash](bool success, const QString &message) {
        emit deleteFinished(stash, success, message);
    });

    connect(stash.data(), &Stash::stashDirChanged, [this,stash](const QString &stashDir) {
        save();
        emit stashDirChanged(stash, stashDir);
    });
    connect(stash.data(), &Stash::mountDirChanged, [this,stash](const QString &mountDir) {
        save();
        emit mountDirChanged(stash, mountDir);
    });
    connect(stash.data(), &Stash::mountOnStartupChanged, [this,stash](bool mountOnStartup) {
        save();
        emit mountOnStartupChanged(stash, mountOnStartup);
    });
    connect(stash.data(), &Stash::savePasswordChanged, [this,stash](bool savePassword) {
        save();
        emit savePasswordChanged(stash, savePassword);
    });
    connect(stash.data(), &Stash::stashGroupChanged, [this,stash](const QString &stashGroup) {
        save();
        emit stashGroupChanged(stash, stashGroup);
    });
    connect(stash.data(), &Stash::showInTrayChanged, [this,stash](bool showInTray) {
        save();
        emit showInTrayChanged(stash, showInTray);
    });
    connect(stash.data(), &Stash::mountWhenAvailableChanged, [this,stash](bool mountWhenAvailable) {
        save();
        emit mountWhenAvailableChanged(stash, mountWhenAvailable);
    });
    connect(stash.data(), &Stash::allowOtherUsersChanged, [this,stash](bool allowOtherUsers) {
        save();
        emit allowOtherUsersChanged(stash, allowOtherUsers);
    });
    connect(stash.data(), &Stash::persistentMountDirectoryChanged, [this,stash](bool persistentMountDirectory) {
        save();
        emit persistentMountDirectoryChanged(stash, persistentMountDirectory);
    });
    connect(stash.data(), &Stash::openFileManagerChanged, [this,stash](bool openFileManager) {
        save();
        emit openFileManagerChanged(stash, openFileManager);
    });
    connect(stash.data(), &Stash::unmountIdleChanged, [this,stash](bool unmountIdle) {
        save();
        emit unmountIdleChanged(stash, unmountIdle);
    });
    connect(stash.data(), &Stash::unmountIdleTimeChanged, [this,stash](int unmountIdleTime) {
        save();
        emit unmountIdleTimeChanged(stash, unmountIdleTime);
    });
    connect(stash.data(), &Stash::encfsConfigFileChanged, [this,stash](const QString &encfsConfigFile) {
        save();
        emit encfsConfigFileChanged(stash, encfsConfigFile);
    });
    connect(stash.data(), &Stash::additionalMountOptionsChanged, [this,stash](const QString &additionalMountOptions) {
        save();
        emit additionalMountOptionsChanged(stash, additionalMountOptions);
    });
    connect(stash.data(), &Stash::preMountCommandChanged, [this,stash](const QString &preMountCommand) {
        save();
        emit preMountCommandChanged(stash, preMountCommand);
    });
    connect(stash.data(), &Stash::postMountCommandChanged, [this,stash](const QString &postMountCommand) {
        save();
        emit postMountCommandChanged(stash, postMountCommand);
    });
    connect(stash.data(), &Stash::preUnmountCommandChanged, [this,stash](const QString &preUnmountCommand) {
        save();
        emit preUnmountCommandChanged(stash, preUnmountCommand);
    });
    connect(stash.data(), &Stash::postUnmountCommandChanged, [this,stash](const QString &postUnmountCommand) {
        save();
        emit postUnmountCommandChanged(stash, postUnmountCommand);
    });
    connect(stash.data(), &Stash::keyringNameChanged, [this,stash](const QString &keyringName) {
        save();
        emit keyringNameChanged(stash, keyringName);
    });
    connect(stash.data(), &Stash::allowNonEmptyMountDirectoryChanged, [this,stash](bool allowNonEmptyMountDirectory) {
        save();
        emit allowNonEmptyMountDirectoryChanged(stash, allowNonEmptyMountDirectory);
    });
    connect(stash.data(), &Stash::allowRootAccessChanged, [this,stash](bool allowRootAccess) {
        save();
        emit allowRootAccessChanged(stash, allowRootAccess);
    });
    connect(stash.data(), &Stash::trayIconIndicateMountedChanged, [this,stash](bool trayIconIndicateMounted) {
        save();
        emit trayIconIndicateMountedChanged(stash, trayIconIndicateMounted);
    });
    connect(stash.data(), &Stash::shortNameChanged, [this,stash](const QString &shortName) {
        emit shortNameChanged(stash, shortName);
    });
    connect(stash.data(), &Stash::isMountedChanged, [this,stash](bool isMounted) {
        emit isMountedChanged(stash, isMounted);
    });
    connect(stash.data(), &Stash::isAvailableChanged, [this,stash](bool isAvailable) {
        emit isAvailableChanged(stash, isAvailable);
    });
    connect(stash.data(), &Stash::isBusyChanged, [this,stash](bool isBusy, const QString &message) {
        emit isBusyChanged(stash, isBusy, message);
    });
    connect(stash.data(), &Stash::isInitializedChanged, [this,stash](bool isInitialized) {
        emit isInitializedChanged(stash, isInitialized);
    });

    m_allStashes.append(stash);

    if (stash->isInitialized()) {
        save();
    }

    m_mutex.unlock();
    rebuildFiltered();
    emit stashAdded(stash);
    return stash;
}

void StashManager::removeStash(QSharedPointer<Stash> &stash)
{
    if (stash->isBusy()) {
        return;
    }

    m_mutex.lock();
    m_allStashes.removeOne(stash);

    stash->disconnect(); // releases the shared pointers to signal handlers

    save();
    m_mutex.unlock();
    rebuildFiltered();
    emit stashRemoved(stash);
}

QList<QSharedPointer<Stash>> StashManager::stashes(const QVariant &stashGroup) const
{
    if (stashGroup.isNull()) {
        return m_allStashes;
    } else {
        return m_filteredStashes.value(stashGroup);
    }
}

int StashManager::size(const QVariant &stashGroup) const
{
    return stashes(stashGroup).size();
}

int StashManager::indexOf(const QVariant &stashGroup, const QSharedPointer<Stash> &stash) const
{
    return stashes(stashGroup).indexOf(stash);
}

QSharedPointer<Stash> StashManager::at(const QVariant &stashGroup, int i) const
{
    if ((i < 0) && (i >= stashes(stashGroup).size())) {
        qDebug() << "Warning: empty stash requested. This is most likely a bug.";
        return QSharedPointer<Stash>::create();
    }

    return stashes(stashGroup).at(i);
}

void StashManager::move(const QVariant &stashGroup, int from, int to)
{
    m_mutex.lock();

    if ( ((from < 0) || (from >= stashes(stashGroup).size())) || ((to < 0) || (to >= stashes(stashGroup).size())) ) {
        return;
    }

    int globalStashIndexFrom = m_allStashes.indexOf(at(stashGroup, from));
    int globalStashIndexTo = m_allStashes.indexOf(at(stashGroup, to));

    m_allStashes.move(globalStashIndexFrom, globalStashIndexTo);

    m_mutex.unlock();
    rebuildFiltered();
    emit orderChanged();
}

QSharedPointer<Stash> StashManager::stashByStashDir(const QString &stashDir)
{
    QString _stashDir = Helpers::pathToAbsolute(stashDir);
    foreach (auto stash, m_allStashes) {
        if ((stash->stashDir() == _stashDir) && (stash->isInitialized())) {
            return stash;
        }
    }
    return QSharedPointer<Stash>();
}

QSharedPointer<Stash> StashManager::stashByMountDir(const QString &mountDir)
{
    QString _mountDir = Helpers::pathToAbsolute(mountDir);
    foreach (auto stash, m_allStashes) {
        if ((stash->mountDir() == _mountDir) && (stash->isInitialized())) {
            return stash;
        }
    }
    return QSharedPointer<Stash>();
}

void StashManager::mount(const QSharedPointer<Stash> &stash, const QString &password)
{
    stash->mount(password);
}

void StashManager::unmount(const QSharedPointer<Stash> &stash)
{
    stash->unmount();
}

void StashManager::create(const QSharedPointer<Stash> &stash, const QString &password, Encfs::CreateStashFlags flags)
{
    stash->create(password, flags);
}

void StashManager::changePassword(const QSharedPointer<Stash> &stash, const QString &oldPassword, const QString &newPassword)
{
    stash->changePassword(oldPassword, newPassword);
}

QString StashManager::info(const QSharedPointer<Stash> &stash) const
{
    return stash->info();
}

void StashManager::cancel(const QSharedPointer<Stash> &stash)
{
    stash->encfs().cancel();
}

void StashManager::delete_(const QSharedPointer<Stash> &stash)
{
    stash->delete_();
}

const QString StashManager::toString()
{
    QString string;
    string.append(QString(QLatin1String("   Mount Directory\t\t\tStash Directory\t\t\tMounted\n")));
    int i=0;
    foreach (auto stash, m_allStashes) {
        string.append(QString(QLatin1String("%1  %2\t\t%3\t\t%4\n")).arg(QString::number(i++)).arg(stash->mountDir()).arg(stash->stashDir()).arg(stash->isMounted()));
    }
    return string;
}

void StashManager::rebuildFiltered()
{
    m_mutex.lock();
    m_filteredStashes.clear();
    foreach (auto stash, m_allStashes) {
        m_filteredStashes[stash->stashGroup()].append(stash);
    }
    m_mutex.unlock();
}

bool StashManager::isBusy() const
{
    foreach (auto stash, m_allStashes) {
        if (stash->isBusy()) {
            return true;
        }
    }
    return false;
}

const QString StashManager::statusMessage()
{
    foreach (auto stash, m_allStashes) {
        if (stash->isBusy()) {
            return stash->statusMessage();
        }
    }
    return QLatin1String();
}

QStringList StashManager::list()
{
    QStringList _list;
    foreach (auto stash, m_allStashes) {
        _list << Helpers::pathToRelative(stash->mountDir());
    }
    return _list;
}

void StashManager::mount(const QString &directory)
{
    QSharedPointer<Stash> stash;

    stash = stashByMountDir(directory);
    if (stash.isNull()) {
        stash = stashByStashDir(directory);
    }

    if (stash.isNull()) {
        return;
    }

    if (stash->isBusy() || stash->isMounted() || !stash->isAvailable()) {
        return;
    }

    mount(stash);

}

void StashManager::unmount(const QString &directory)
{
    QSharedPointer<Stash> stash;

    stash = stashByMountDir(directory);
    if (stash.isNull()) {
        stash = stashByStashDir(directory);
    }

    if (stash.isNull()) {
        return;
    }

    if (stash->isBusy() || !stash->isMounted()) {
        return;
    }

    unmount(stash);
}

void StashManager::refresh()
{
    foreach (auto stash, m_allStashes) {
        stash->refresh();
    }
}

void StashManager::start()
{
    //m_mountSurveiler->start();
}

void StashManager::stop()
{
   //m_mountSurveiler->stop();
}

void StashManager::load()
{
    foreach (auto stash, m_allStashes) {
        stash->disconnect();
    }

    m_allStashes.clear();

    QSettings settings;
    settings.setIniCodec("UTF-8");
    int size = settings.beginReadArray("Stashes");
    for (int i = 0; i < size; i++) {
        settings.setArrayIndex(i);
        QSharedPointer<Stash> stash = addStash(QSharedPointer<Stash>::create(settings));
        stash->setIsInitialized(true);
    }
    settings.endArray();
    rebuildFiltered();
}

void StashManager::save()
{
    QSettings settings;
    settings.setIniCodec("UTF-8");
    settings.beginWriteArray("Stashes");
    settings.remove("");
    for (int i = 0; i < m_allStashes.size(); i++) {
        settings.setArrayIndex(i);
        if (m_allStashes.at(i)->isInitialized()) {
            m_allStashes.at(i)->save(settings);
        }
    }
    settings.endArray();
}

void StashManager::startup()
{
    foreach (auto stash, m_allStashes) {
        if (stash->isAvailable() && !stash->isMounted() && stash->mountOnStartup()) {
            QTimer::singleShot(0, [this, stash]() {
                mount(stash);
            });
        }
    }
}

void StashManager::shutdown()
{
    m_mountSurveiler.reset();
    save();

    if (Settings::instance().unmountAllOnExit()) {
        bool mounted = false;
        foreach (auto stash, m_allStashes) {
            if (stash->isMounted()) {
                mounted = true;
                break;
            }
        }

        if (!mounted) {
            QCoreApplication::instance()->quit();
            return;
        }


        foreach (auto stash, m_allStashes) {
            QTimer::singleShot(0, [this, stash]() {
                unmount(stash);
            });
        }

        connect(this, &StashManager::unmountFinished, [this]() {
            foreach (auto stash, m_allStashes) {
                if (stash->isMounted()) {
                    return;
                }
            }
            QCoreApplication::instance()->quit();
        });

    } else {
        QCoreApplication::instance()->quit();
    }
}
