#ifndef STASHMANAGER_H
#define STASHMANAGER_H

// Qt
#include <QMap>
#include <QMutex>
#include <QObject>
#include <QSharedPointer>
#include <QVariant>

#include "encfs.h"

class MUnixMountSurveiler;
class Stash;

class StashManager : public QObject
{
    Q_OBJECT

public:
    explicit StashManager(QObject *parent = nullptr);
    ~StashManager();

    static StashManager &instance();

    QSharedPointer<Stash> addStash(const QSharedPointer<Stash> &stash);
    void removeStash(QSharedPointer<Stash> &stash);

    void mount(const QSharedPointer<Stash> &stash, const QString &password = QLatin1String());
    void unmount(const QSharedPointer<Stash> &stash);
    void create(const QSharedPointer<Stash> &stash, const QString &password, Encfs::CreateStashFlags flags = 0);
    void changePassword(const QSharedPointer<Stash> &stash, const QString &oldPassword, const QString &newPassword);
    QString info(const QSharedPointer<Stash> &stash) const;
    void cancel(const QSharedPointer<Stash> &stash);
    void delete_(const QSharedPointer<Stash> &stash);

    /*!
     * when stashGroup is QVariant(), it means all stashes,
     * "" is the default (no group set), else the stashgroup name as QString
     */
    QList<QSharedPointer<Stash>> stashes(const QVariant &stashGroup = QVariant()) const;
    int size(const QVariant &stashGroup = QVariant()) const;
    int indexOf(const QVariant &stashGroup, const QSharedPointer<Stash> &stash) const;
    QSharedPointer<Stash> at(const QVariant &stashGroup, int i) const;
    void move(const QVariant &stashGroup, int from, int to);

    int stashGroupId(const QVariant &stashGroup) const;

    QSharedPointer<Stash> stashByStashDir(const QString &stashDir);
    QSharedPointer<Stash> stashByMountDir(const QString &mountDir);

    const QString toString();
    bool isBusy() const;
    const QString statusMessage();

signals:
    void stashesChanged();
    void orderChanged();
    void stashAdded(const QSharedPointer<Stash> &stash);
    void stashRemoved(const QSharedPointer<Stash> &stash);

public slots:
    QStringList list();
    void mount(const QString &directory);
    void unmount(const QString &directory);

    void refresh();
    void start();
    void stop();
    void load();
    void save();
    void startup();
    void shutdown();

private:
    void rebuildFiltered();

private:
    static StashManager *m_instance;
    QMutex m_mutex;
    QList< QSharedPointer<Stash> > m_allStashes;
    QMap< QSharedPointer<Stash>, QList<QMetaObject::Connection> > m_connections;
    QMap< QVariant, QList<QSharedPointer<Stash>> > m_filteredStashes;
    QScopedPointer<MUnixMountSurveiler> m_mountSurveiler;


// Signals from stashes ===========================================================================
signals:
    void mountFinished(const QSharedPointer<Stash> &stash, bool success, const QString &message = QLatin1String());
    void unmountFinished(const QSharedPointer<Stash> &stash, bool success, const QString &message = QLatin1String());
    void createFinished(const QSharedPointer<Stash> &stash, bool success, const QString &message = QLatin1String());
    void changePasswordFinished(const QSharedPointer<Stash> &stash, bool success, const QString &message = QLatin1String());
    void deleteFinished(const QSharedPointer<Stash> &stash, bool success, const QString &message = QLatin1String());

    void stashDirChanged(const QSharedPointer<Stash> &stash, const QString &stashDir);
    void mountDirChanged(const QSharedPointer<Stash> &stash, const QString &mountDir);
    void mountOnStartupChanged(const QSharedPointer<Stash> &stash, bool mountOnStartup);
    void savePasswordChanged(const QSharedPointer<Stash> &stash, bool savePassword);
    void stashGroupChanged(const QSharedPointer<Stash> &stash, const QString &stashGroup);
    void showInTrayChanged(const QSharedPointer<Stash> &stash, bool showInTray);
    void mountWhenAvailableChanged(const QSharedPointer<Stash> &stash, bool mountWhenAvailable);
    void allowOtherUsersChanged(const QSharedPointer<Stash> &stash, bool allowOtherUsers);
    void persistentMountDirectoryChanged(const QSharedPointer<Stash> &stash, bool persistentMountDirectory);
    void openFileManagerChanged(const QSharedPointer<Stash> &stash, bool openFileManager);
    void unmountIdleChanged(const QSharedPointer<Stash> &stash, bool unmountIdle);
    void unmountIdleTimeChanged(const QSharedPointer<Stash> &stash, int unmountIdleTime);
    void encfsConfigFileChanged(const QSharedPointer<Stash> &stash, const QString &encfsConfigFile);
    void additionalMountOptionsChanged(const QSharedPointer<Stash> &stash, const QString &additionalMountOptions);
    void preMountCommandChanged(const QSharedPointer<Stash> &stash, const QString &preMountCommand);
    void postMountCommandChanged(const QSharedPointer<Stash> &stash, const QString &postMountCommand);
    void preUnmountCommandChanged(const QSharedPointer<Stash> &stash, const QString &preUnmountCommand);
    void postUnmountCommandChanged(const QSharedPointer<Stash> &stash, const QString &postUnmountCommand);
    void keyringNameChanged(const QSharedPointer<Stash> &stash, const QString &keyringName);
    void allowNonEmptyMountDirectoryChanged(const QSharedPointer<Stash> &stash, bool allowNonEmptyMountDirectory);
    void allowRootAccessChanged(const QSharedPointer<Stash> &stash, bool allowRootAccess);
    void trayIconIndicateMountedChanged(const QSharedPointer<Stash> &stash, bool trayIconIndicateMounted);
    void shortNameChanged(const QSharedPointer<Stash> &stash, const QString &shortName);
    void isMountedChanged(const QSharedPointer<Stash> &stash, bool isMounted);
    void isAvailableChanged(const QSharedPointer<Stash> &stash, bool isaAvailable);
    void isBusyChanged(const QSharedPointer<Stash> &stash, bool isBusy, const QString &message = QLatin1String());
    void isInitializedChanged(const QSharedPointer<Stash> &stash, bool isInitialized);

    friend class Stash;
};

#endif // STASHMANAGER_H
