#include <signal.h>

// Qt
#include <QApplication>
#include <QCommandLineOption>
#include <QCommandLineParser>
#include <QDBusConnection>
#include <QDebug>
#include <QIcon>
#include <QLoggingCategory>
#include <QSettings>
#include <QStyle>
#include <QTextStream>
#include <QTimer>

// Core
#include "Core/DBus/stashmanager_adaptor.h"
#include "Core/DBus/stashmanager_interface.h"
#include "Core/settings.h"
#include <Core/stashmanager.h>
#include <Core/stash.h>

#include "application.h"
#include "stashtableview.h"
#include "managerwindow.h"
#include "mounter.h"
#include "trayicon.h"

#include "DBus/managerwindow_adaptor.h"
#include "DBus/managerwindow_interface.h"

com::moritzmolch::EncfsManager::StashManager *stashManagerInterface()
{
    QDBusConnection dbusConnection = QDBusConnection::sessionBus();
    if (!dbusConnection.isConnected()) {
        qDebug() << "Failed to connect to DBus session.";
        exit(0);
    }

    return new com::moritzmolch::EncfsManager::StashManager (
                                  QLatin1String("com.moritzmolch.EncfsManager"),
                                  QLatin1String("/com/moritzmolch/EncfsManager/StashManager"),
                                  QDBusConnection::sessionBus());
}

void processArguments(const QGuiApplication &app)
{
    QCommandLineParser parser;
    parser.addHelpOption();
    parser.addVersionOption();

    QCommandLineOption listOption(QStringList() << "list",
            QCoreApplication::translate("main", "Lists all stashes."));
    parser.addOption(listOption);

    QCommandLineOption mountOption(QStringList() << "mount",
            QCoreApplication::translate("main", "Mounts the stash."),
            QCoreApplication::translate("main", "stash"));
    parser.addOption(mountOption);

    QCommandLineOption unmountOption(QStringList() << "unmount",
            QCoreApplication::translate("main", "Unmounts the stash."),
            QCoreApplication::translate("main", "stash"));
    parser.addOption(unmountOption);

    parser.process(app);

    if (parser.isSet(listOption)) {
        QDBusPendingReply<QStringList> reply = stashManagerInterface()->list();
        reply.waitForFinished();

        QTextStream out(stdout);
        out << reply.value().join("\n") << endl;

        exit(0);

    } else if (parser.isSet(mountOption)) {
        QDBusPendingReply<QString> reply = stashManagerInterface()->mount(parser.value(mountOption));
        reply.waitForFinished();

        QTextStream out(stdout);
        out << reply.value() << endl;

        exit(0);

    } else if (parser.isSet(unmountOption)) {
        QDBusPendingReply<QString> reply = stashManagerInterface()->unmount(parser.value(unmountOption));
        reply.waitForFinished();

        QTextStream out(stdout);
        out << reply.value() << endl;

        exit(0);
    }
}

#define STRINGIFY(x) #x
#define TOSTRING(x) STRINGIFY(x)

int main(int argc, char *argv[])
{
    QLoggingCategory::setFilterRules("*=false");
    QCoreApplication::setAttribute(Qt::AA_EnableHighDpiScaling);
    QCoreApplication::setAttribute(Qt::AA_UseHighDpiPixmaps);
    QCoreApplication::setAttribute(Qt::AA_DontUseNativeMenuBar); // Native menu causes problems on Ubuntu 17.10 with Gnome Shell
    QCoreApplication::setOrganizationName("Moritz Molch");
    QCoreApplication::setOrganizationDomain("moritzmolch.com");
    QCoreApplication::setApplicationName("Molch Encfs Manager");
    QCoreApplication::setApplicationVersion(TOSTRING(VERSION));
    QGuiApplication::setApplicationDisplayName("Molch Encfs Manager");


    Application app(argc, argv);
    app.setQuitOnLastWindowClosed(false);

    processArguments(app);

    QDBusConnection dbusConnection = QDBusConnection::sessionBus();
    if (!dbusConnection.isConnected()) {
        qDebug() << "Failed to connect to DBus session. The application might not work as expected.";
    }

    if (!dbusConnection.registerService(QLatin1String("com.moritzmolch.EncfsManager"))) {
        com::moritzmolch::EncfsManager::ManagerWindow managerWindowInterface(
                                      QLatin1String("com.moritzmolch.EncfsManager"),
                                      QLatin1String("/com/moritzmolch/EncfsManager/ManagerWindow"),
                                      QDBusConnection::sessionBus());
        managerWindowInterface.setVisible(true);
        return 0;
    }


    Settings settings; Q_UNUSED(settings)

    //new SettingsAdaptor(&settings);
    //dbusConnection.registerObject("/com/moritzmolch/EncfsManager/Settings", &settings);

    StashManager stashManager; Q_UNUSED(stashManager)

    new StashManagerAdaptor(&stashManager);
    dbusConnection.registerObject("/com/moritzmolch/EncfsManager/StashManager", &stashManager);

    QIcon::setThemeName("encfsmanager");
    //QApplication::setWindowIcon(QIcon::fromTheme(QLatin1String("gnome-encfs-manager-panel")));
    QApplication::setWindowIcon(QIcon(QLatin1String(":/icons/encfsmanager/96/gnome-encfs-manager-panel.png")));

    //QFont font = app.font();
    //font.setPointSize(9);
    //app.setFont(font);

    app.setStyle("fusion");
    app.setStyleSheet("QGroupBox {"
                      "font-weight: bold;"
                      "background-color: rgba(0, 0, 0, 0);"
                      "border: 0;"
                      "color: rgb(50, 50, 50);"
                      "margin-top: 2ex; /* leave space at the top for the title */"
                      "}");


    Mounter mounter; Q_UNUSED(mounter)
    ManagerWindow managerWindow;
    new ManagerWindowAdaptor(&managerWindow);
    dbusConnection.registerObject("/com/moritzmolch/EncfsManager/ManagerWindow", &managerWindow);

    TrayIcon trayIcon; Q_UNUSED(trayIcon)

    if (!(Settings::instance().trayIconEnabled() && Settings::instance().trayIconStartHidden()) || Settings::instance().isFirstRun()) {
        managerWindow.show();
    }

    /*if (Settings::instance().trayIconStartHidden() && !TrayIcon::isAvailable()) {
        managerWindow.showMinimized();
    }*/

    StashManager::instance().startup();

    struct sigaction saQuit = {0};
    saQuit.sa_handler = [](int) {
        StashManager::instance().shutdown();
        //QCoreApplication::instance()->quit();
    };
    sigaction(SIGQUIT, &saQuit, NULL);

    struct sigaction saTerm = {0};
    saTerm.sa_handler = [](int) {
        StashManager::instance().shutdown();
        //QCoreApplication::instance()->quit();
    };
    sigaction(SIGTERM, &saTerm, NULL);

    int exitCode = app.exec();

    Settings::instance().setIsFirstRun(false);

    return exitCode;
}
