#include "managerwindow.h"
#include "ui_managerwindow.h"

//Qt
#include <QCloseEvent>
#include <QDebug>
#include <QDesktopServices>
#include <QLabel>
#include <QMenu>
#include <QScreen>
#include <QSettings>

// Core
#include "Core/settings.h"
#include <Core/stash.h>
#include <Core/stashmanager.h>

#include <MAboutDialog>
#include <MHelpBrowser>
#include <MSpinner>
#include "pager.h"
#include "pages/addstashpage.h"
#include "pages/changepasswordpage.h"
#include "pages/configurestashpage.h"
#include "pages/configurestashexpertpage.h"
#include "pages/mountstashpage.h"
#include "pages/preferencespage.h"
#include "pages/preferencesexpertpage.h"
#include "pages/removestashpage.h"
#include "pages/stashcommandpage.h"
#include "pages/stashinformationpage.h"
#include "stashtableview.h"
#include "trayicon.h"

ManagerWindow *ManagerWindow::m_instance = nullptr;

ManagerWindow::ManagerWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::ManagerWindow),
    m_popupmenu(new QMenu(this))
{
    if (m_instance != nullptr) {
        qWarning() << "Only one instance of ManagerWindow is allowed";
        return;
    }
    m_instance = this;
    ui->setupUi(this);

    bool isBusy = StashManager::instance().isBusy();

    QWidget *toolBarSpacer = new QWidget(this);
    toolBarSpacer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    ui->toolBar->addWidget(toolBarSpacer);

    MSpinner *spinner = new MSpinner();
    spinner->setMinimumSize(20, 20);
    spinner->setVisible(isBusy);
    QAction *action_spinner = ui->toolBar->addWidget(spinner);

    QLabel *label_statusMessage = new QLabel;
    label_statusMessage->setMinimumHeight(18);
    label_statusMessage->setContentsMargins(4, 0, 0, 0);
    ui->statusbar->addWidget(label_statusMessage);

    if (isBusy) {
        label_statusMessage->setText(StashManager::instance().statusMessage());
    } else {
        label_statusMessage->setText(tr("You can rearrange the items via drag and drop"));
    }

    connect(&StashManager::instance(), &StashManager::isBusyChanged, [action_spinner, label_statusMessage]() {
        bool isBusy = StashManager::instance().isBusy();

        action_spinner->setVisible(isBusy);
        if (isBusy) {
            label_statusMessage->setText(StashManager::instance().statusMessage());
        } else {
            label_statusMessage->setText(tr("You can rearrange the items via drag and drop"));
        }
    });

    QWidget *toolBarSpacer2 = new QWidget(this);
    toolBarSpacer2->setMinimumWidth(6);
    ui->toolBar->addWidget(toolBarSpacer2);

    ui->toolBar->addAction(ui->actionMenu);

    ui->actionMount->setEnabled(false);
    ui->actionUnmount->setEnabled(false);
    ui->actionChangePassword->setEnabled(false);
    ui->actionRemoveStash->setEnabled(false);
    ui->actionShowInformation->setEnabled(false);
    ui->actionConfigureStash->setEnabled(false);
    connect(ui->stashTableView, &StashTableView::selectedStashChanged, [this](const QSharedPointer<Stash> stash) {
        if (stash.isNull()) {
            ui->actionMount->setEnabled(false);
            ui->actionUnmount->setEnabled(false);
            ui->actionChangePassword->setEnabled(false);
            ui->actionRemoveStash->setEnabled(false);
            ui->actionShowInformation->setEnabled(false);
            ui->actionConfigureStash->setEnabled(false);
            return;
        }

        //if (!stash->isEmpty()) {
            ui->actionRemoveStash->setEnabled(true);
            ui->actionChangePassword->setEnabled(stash->isAvailable());
            ui->actionShowInformation->setEnabled(stash->isAvailable());
            ui->actionConfigureStash->setEnabled(true);
            ui->actionMount->setEnabled(!stash->isMounted() && stash->isAvailable());
            ui->actionUnmount->setEnabled(stash->isMounted());
        //} else {
        //    ui->actionRemoveStash->setEnabled(false);
        //    ui->actionConfigureStash->setEnabled(false);
        //}
    });

    ui->actionStashGroups->setChecked(Settings::instance().managerShowStashGroups());
    ui->frame_stashgroups->setVisible(Settings::instance().managerShowStashGroups());
    connect(ui->actionStashGroups, &QAction::triggered, &Settings::instance(), &Settings::setManagerShowStashGroups);

    if (Settings::instance().managerShowStashGroups()) {
        ui->frame_stashTable->setStyleSheet(".QFrame#frame_stashTable {"
                                            "border: 0;"
                                            "border-bottom: 1px solid #b8b8b8;"
                                            "border-left: 1px solid #b8b8b8;"
                                            "}");
    } else {
        ui->frame_stashTable->setStyleSheet(".QFrame#frame_stashTable {"
                                            "border: 0;"
                                            "border-bottom: 1px solid #b8b8b8;"
                                            "}");
    }

    connect(&Settings::instance(), &Settings::managerShowStashGroupsChanged, [this](bool managerShowStashGroups) {
        ui->actionStashGroups->setChecked(managerShowStashGroups);
        ui->frame_stashgroups->setVisible(managerShowStashGroups);
        if (managerShowStashGroups) {
            ui->frame_stashTable->setStyleSheet(".QFrame#frame_stashTable {"
                                                "border: 0;"
                                                "border-bottom: 1px solid #b8b8b8;"
                                                "border-left: 1px solid #b8b8b8;"
                                                "}");
        } else {
            ui->frame_stashTable->setStyleSheet(".QFrame#frame_stashTable {"
                                                "border: 0;"
                                                "border-bottom: 1px solid #b8b8b8;"
                                                "}");
        }
    });


    ui->menuBar->hide();
    m_popupmenu->addMenu(ui->menuManager);
    m_popupmenu->addMenu(ui->menuStash);
    m_popupmenu->addMenu(ui->menuView);
    m_popupmenu->addMenu(ui->menuHelp);
    //menu->adjustSize();

    m_popupmenu->adjustSize();
    connect(m_popupmenu, &QMenu::aboutToHide, [=]() {
        ui->actionMenu->setChecked(false);
    });

    connect(ui->stashGroupWidget, &StashgroupWidget::selectedStashGroupChanged, [this](const QVariant &stashGroup) {
        ui->stashTableView->setStashGroup(stashGroup);
    });

    //connect(&StashManager::instance, &StashManager::mou)

    TrayIcon::update();
}

ManagerWindow::~ManagerWindow()
{
    if (isVisible()) {
        QSettings settings;
        settings.beginGroup(QLatin1String("ManagerWindow"));
        settings.setValue("geometry", saveGeometry());
        settings.setValue("windowState", saveState());
        settings.endGroup();
    }

    delete ui;
    m_instance = nullptr;
    TrayIcon::update();
}

ManagerWindow *ManagerWindow::instance()
{
    return m_instance;
}

void ManagerWindow::on_actionSettings_triggered()
{
    Pager pager(tr("Preferences"), this);
    pager.addPage(new PreferencesPage);
    pager.addPage(new PreferencesExpertPage);
    pager.selectPage(0);
    pager.adjustSize();
    pager.exec();
}

void ManagerWindow::on_actionConfigureStash_triggered()
{
    auto stash = ui->stashTableView->selectedStash();
    Pager pager(stash, this);
    pager.addPage(new ConfigureStashPage(stash));
    pager.addPage(new ConfigureStashExpertPage(stash));
    pager.addPage(new StashCommandPage(stash));
    pager.addPage(new ChangePasswordPage(stash));
    pager.addPage(new StashInformationPage(stash));
    pager.addPage(new RemoveStashPage(stash));
    pager.selectPage(0);
    pager.adjustSize();
    pager.exec();
}

void ManagerWindow::on_actionShowInformation_triggered()
{
    auto stash = ui->stashTableView->selectedStash();
    Pager pager(stash, this);
    pager.addPage(new StashInformationPage(stash));
    pager.insertPage(0, new ChangePasswordPage(stash));
    pager.insertPage(0, new StashCommandPage(stash));
    pager.insertPage(0, new ConfigureStashExpertPage(stash));
    pager.insertPage(0, new ConfigureStashPage(stash));
    pager.addPage(new RemoveStashPage(stash));
    pager.selectPage(4);
    pager.adjustSize();
    pager.exec();
}

void ManagerWindow::on_actionRefreshStashes_triggered()
{
    StashManager::instance().refresh();
}

void ManagerWindow::on_actionAddStash_triggered()
{
    QSharedPointer<Stash> stash(new Stash);
    StashManager::instance().addStash(stash);
    Pager pager(stash, this);
    pager.addPage(new AddStashPage(stash));
    pager.addPage(new ConfigureStashPage(stash));
    pager.addPage(new ConfigureStashExpertPage(stash));
    pager.addPage(new StashCommandPage(stash));
    pager.selectPage(0);
    pager.adjustSize();

    if (pager.exec() == QDialog::Rejected) {
        StashManager::instance().removeStash(stash);
    }
}

void ManagerWindow::on_actionRemoveStash_triggered()
{
    auto stash = ui->stashTableView->selectedStash();
    Pager pager(stash, this);
    pager.addPage(new RemoveStashPage(stash));
    pager.insertPage(0, new StashInformationPage(stash));
    pager.insertPage(0, new ChangePasswordPage(stash));
    pager.insertPage(0, new StashCommandPage(stash));
    pager.insertPage(0, new ConfigureStashExpertPage(stash));
    pager.insertPage(0, new ConfigureStashPage(stash));
    pager.selectPage(5);
    pager.adjustSize();
    pager.exec();
}

void ManagerWindow::on_actionChangePassword_triggered()
{
    auto stash = ui->stashTableView->selectedStash();
    Pager pager(stash, this);
    pager.addPage(new ChangePasswordPage(stash));
    pager.insertPage(0, new StashCommandPage(stash));
    pager.insertPage(0, new ConfigureStashExpertPage(stash));
    pager.insertPage(0, new ConfigureStashPage(stash));
    pager.addPage(new StashInformationPage(stash));
    pager.addPage(new RemoveStashPage(stash));
    pager.selectPage(3);
    pager.adjustSize();
    pager.exec();
}

void ManagerWindow::on_actionMenu_triggered()
{
    m_popupmenu->move(mapToGlobal(QPoint(width()-m_popupmenu->width(),ui->toolBar->height()-2)));
    m_popupmenu->exec();
}

void ManagerWindow::on_actionHelp_triggered()
{
    MHelpBrowser::open();
}

void ManagerWindow::on_actionAbout_triggered()
{
    MAboutDialog dlg(this);
    dlg.setHomepage(QLatin1String("https://moritzmolch.com/apps/mencfsm"));
    dlg.setLicenseHelpPage(QLatin1String("license.html"));
    dlg.setAuthorsHelpPage(QLatin1String("authors.html"));
    dlg.setUpdateUrl(QUrl(QStringLiteral("http://moritzmolch.com/apps/download/mencfsm/version.json")));
    dlg.exec();
}

void ManagerWindow::on_actionAbout_Qt_triggered()
{
    QApplication::aboutQt();
}

void ManagerWindow::on_actionQuit_triggered()
{
    StashManager::instance().shutdown();
}

void ManagerWindow::on_actionClose_triggered()
{
    close();
}

void ManagerWindow::setVisible(bool visible)
{
    if (isVisible() == visible) {
        // The window won't show above other windows otherwise
        // Might cause problem with window show animation from minimized
        // Testing required
        QMainWindow::setVisible(false);
        QMainWindow::setVisible(visible);
        activateWindow();
        raise();
        return;
    }

    if (visible) {
        QSettings settings;
        settings.beginGroup(QLatin1String("ManagerWindow"));
        restoreGeometry(settings.value("geometry").toByteArray());
        restoreState(settings.value("windowState").toByteArray());
        settings.endGroup();
    } else {
        QSettings settings;
        settings.beginGroup(QLatin1String("ManagerWindow"));
        settings.setValue("geometry", saveGeometry());
        settings.setValue("windowState", saveState());
        settings.endGroup();
    }

    if (visible) {
        QSize screenSize = QApplication::screens().at(0)->size();
        move(screenSize.width()/2 - width()/2,
             screenSize.height()/2 - height()/2);
    }

    QMainWindow::setVisible(visible);
    activateWindow();
    raise();
    TrayIcon::update();
}


void ManagerWindow::closeEvent(QCloseEvent *event)
{
    QSettings settings;
    settings.beginGroup(QLatin1String("ManagerWindow"));
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.endGroup();

    event->ignore();
    if (Settings::instance().trayIconEnabled()) {
        setVisible(false);
        //deleteLater();
    } else {
        StashManager::instance().shutdown();
    }
}

void ManagerWindow::on_actionDonate_triggered()
{
    QDesktopServices::openUrl(QStringLiteral("https://moritzmolch.com/apps/Donate.html"));
}
